
import unicodedata
import re
import hashlib
import random
import logging

# Taken from Django
def slugify(value, allow_unicode=False):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    return re.sub(r'[-\s]+', '-', value)

def generate_nonce():
    """str: Create a nonce."""
    return hashlib.md5(str(random.random()).encode()).hexdigest()

def print_error(error):
    LOGGER.error(error)

    if hasattr(error, "__cause__") and error.__cause__ is not None:
        LOGGER.error(error.__cause__)

LOGGER = logging.getLogger(__name__)




import logging
logger = logging.getLogger('StegoVeritas:Modules:Multi:Analysis:Exif')

import os
import random
import hashlib
import json
import subprocess
from prettytable import PrettyTable

from ....helpers import slugify

def run(multi):
    """Checks for any Exif data in/about the file.

    Args:
        image: MultiHandler class instance

    Returns:
        None

    Saves the result to RESULTSDIR/exif/
    """

    args = multi.veritas.args

    # Nothing to do
    if not args.auto and not args.exif:
        logger.debug('Nothing to do.')
        return

    # Note: Python3.6+ added json.loads support for bytes. Thus using decode for backwards compatibility.
    exif = json.loads(subprocess.check_output(['exiftool', '-j', '-b', multi.veritas.file_name ]).decode())

    table = PrettyTable(['key','value'])
    table.align='l'
    exif_values = []

    for key, value in exif[0].items():
        table.add_row([key, value])
        exif_values.append((key, value))

    print("Exif\n====")
    print(table)

    # Save it out
    save_dir = os.path.join(multi.veritas.results_directory, 'exif')

    os.makedirs(save_dir, exist_ok=True)

    for key,value in exif_values:
        if type(value) in [int, float, bool]:
            value = str(value)

        outfile = os.path.join(save_dir, slugify(key))

        # Slugify means that we might have collisions
        if os.path.exists(outfile):
            logger.warn('Exif outpat already exists, modifying.')
            outfile += '_' + hashlib.md5(str(random.random()).encode()).hexdigest()

        if isinstance(value, list):
            value = ", ".join(str(x) for x in value)

        with open(outfile, 'wb') as f:
            f.write(value.encode())





import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:BruteLSB')

import os
import shutil
import multiprocessing
from ....helpers import print_error

def run_dump(image, red_index=None, green_index=None, blue_index=None, alpha_index=None):
    logger.debug("Trying red={} green={} blue={} alpha={}".format(red_index, green_index, blue_index, alpha_index))
    o = image.dumpLSBRGBA(red_index=red_index, green_index=green_index, blue_index=blue_index, alpha_index=alpha_index)
    image.veritas.test_output(o)

def run(image):
    """Attempts to brute force out any LSB stego.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR
    """

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.bruteLSB:
        logger.debug('Nothing to do.')
        return

    if image.file.mode == "P":
        logger.debug('Image is colormap, skipping..')
        return

    with multiprocessing.Pool() as pool:
        wait_for = []

        # Try all the same indexes first. More likely stuff up front.
        # i.e.: 0,0,0  1,1,1,  2,2,2
        for i in range(0,8):
            wait_for.append(pool.apply_async(run_dump, args=(image,), kwds={'red_index': [i], "green_index": [i], "blue_index": [i]}, error_callback=print_error))

        # Try Red
        i = []
        for x in range(8):
            i.append(x)
            wait_for.append(pool.apply_async(run_dump, args=(image,), kwds={'red_index': i[::]}, error_callback=print_error))

        # Try Green
        i = []
        for x in range(8):
            i.append(x)
            wait_for.append(pool.apply_async(run_dump, args=(image,), kwds={'green_index': i[::]}, error_callback=print_error))

        # Try Blue
        i = []
        for x in range(8):
            i.append(x)
            wait_for.append(pool.apply_async(run_dump, args=(image,), kwds={'blue_index': i[::]}, error_callback=print_error))

        # Try Alpha
        if "A" in image.file.mode:
            i = []
            for x in range(8):
                i.append(x)
                wait_for.append(pool.apply_async(run_dump, args=(image,), kwds={'alpha_index': i[::]}, error_callback=print_error))

        # Try across the board
        # i.e.: 0,0,0  01,01,01 etc
        i = []
        for x in range(8):
            i.append(x)
            wait_for.append(pool.apply_async(run_dump, args=(image,), kwds={'red_index': i[::], "green_index": i[::], "blue_index": i[::]}, error_callback=print_error))

        # Wait for everything to finish
        for w in wait_for: w.wait()

    return
    for r in range(0,8):
        for g in range(0,8):
            for b in range(0,8):
                logger.debug("Trying {0}.{1}.{2}".format(r,g,b))
                o = image.dumpLSBRGBA(red_index=[r],green_index=[g],blue_index=[b])
                image.veritas.test_output(o)




import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:ColorMap')

import os
import numpy
from copy import copy

def run(image):
    """Performs colormap stego extraction.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR
    """

    args = image.veritas.args

    # Nothing to do
    if not args.auto and args.colorMap is None and args.colorMapRange is None:
        logger.debug('Nothing to do.')
        return

    extract_colormap(image)

def extract_colormap(image, saveMap=None):

    bands = image.file.getbands()

    if "P" not in bands:
        logger.debug('Not a colormap. Skipping.')
        return

    if image.veritas.args.colorMap in [[], None]:
        index_to_save = range(0x100)
    else:
        index_to_save = image.veritas.args.colorMap

    if saveMap is None:
        if image.veritas.args.colorMapRange is not None:
            saveMap = range(*image.veritas.args.colorMapRange)
        else:
            saveMap = []

    logger.debug('Extracting colorMap. saveMap={}'.format(saveMap))

    # f.putpalette(f.palette.getdata()[1])
    pal = [255 for x in range(768)]

    for save in saveMap:
            pal[save*3] = 0
            pal[save*3+1] = 0
            pal[save*3+2] = 0
    
    # Break apart the color map to help see things
    #for i in range(0xff + 1):
    for i in index_to_save:
            pal2 = copy(pal)
            pal2[i*3] = 0
            pal2[i*3+1] = 0
            pal2[i*3+2] = 0
            image.file.putpalette(pal2)
            # Change to numpy
            #g = numpy.array(f)
            # Xor our colormap index (keep our save here)
            #g = g ^ (i | saveMap)
            # Contrast
            #g[g != 0] = 0xff
            # Revert to image
            #g = Image.fromarray(g)
            # Save
            image.file.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_{0}.png".format(i)))




import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:StegHide')

import os
import hashlib
import random
import subprocess

def run(image):
    """Attempts to extract steghide based stego.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR/steghide.bin
    """

    global output_file
    output_file = os.path.join(image.veritas.results_directory, "steghide.bin")

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.steghide:
        logger.debug('Nothing to do.')
        return

    # Always try without password
    try_extract(image)

    # Try password
    if args.password is not None:
        try_extract(image, args.password)

    # Brute force wordlist
    if args.wordlist is not None:

        with open(args.wordlist, "r") as f:
            line = f.readline().strip()

            while line:
                try_extract(image, line)
                line = f.readline().strip()

def try_extract(image, password=None):

    nonce = hashlib.md5(str(random.random()).encode()).hexdigest()
    outfile = os.path.join(image.veritas.results_directory, "steghide_" + nonce + ".bin")

    try:

        subprocess.check_output(["steghide", "extract", "-sf", image.veritas.file_name, "-xf", outfile, "-p", password or ""], stderr=subprocess.PIPE)
        print("Found something with StegHide: " + outfile)

    except:
        pass





import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:Filters')

import os
from PIL import Image, ImageFilter, ImageFile, ImageEnhance, ImageOps
import numpy
from copy import copy
import multiprocessing
from ....helpers import print_error

def run_filter(image, f_a, filt):
    try:
        g = f_a.filter(filt)
    except IOError as e:
        # Generically print it, then try again (truncated error)
        logger.error("%s", e)
        ImageFile.LOAD_TRUNCATED_IMAGES = True
        g = f_a.filter(filt)

    g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_" + filt.name.replace(" ","_") + ".png"))

def run_enhancer(image, enhancers, outname):
    img = image.file

    # Iteratively apply enhancements
    for enhancer, adjustment in enhancers:
        img = enhancer(img).enhance(adjustment)

    img.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_" + outname))

def run_image_op(image, op, outname):
    img = op(image.file)
    img.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_" + outname))

def run_invert(image, f_a):
    g = numpy.array(f_a)
    g = g ^ 0xff
    g = Image.fromarray(g)
    g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_inverted.png"))

def run_color_planes(image, f_a, plane):

    if plane == "red":
        g = numpy.array(f_a)
        g[:,:,2] *= 0
        g[:,:,1] *= 0
        g = Image.fromarray(g)
        g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_red_plane.png"))
    
    elif plane == "green":
        g = numpy.array(f_a)
        g[:,:,2] *= 0
        g[:,:,0] *= 0
        g = Image.fromarray(g)
        g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_green_plane.png"))

    elif plane == "blue":
        g = numpy.array(f_a)
        g[:,:,0] *= 0
        g[:,:,1] *= 0
        g = Image.fromarray(g)
        g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_blue_plane.png"))

    elif plane == "alpha":
        # TODO: Still assuming all the positions for the given colors in the tuple
        if "A" in image.file.getbands():
                g = numpy.array(image.file)
                g[:,:,0] *= 0
                g[:,:,1] *= 0
                g[:,:,2] *= 0
                g = Image.fromarray(g)
                g.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_alpha_plane.png"))


def run_bit_plane(image, f_a, plane, color, index):
    """Extracts a specific bit plane from the image (i.e.: "Green" bit 0)"""

    temp = numpy.array(plane)
    # Shift over to get right offset
    temp = temp >> index
    # Get the bit
    temp = temp & 1
    # Fix-up the background
    temp[temp == 1] = 255
    # Save it off
    temp = Image.fromarray(temp)
    temp.save(os.path.join(image.veritas.results_directory, os.path.basename(image.veritas.file_name) + "_{0}_{1}.png".format(color,index)))

def run(image):
    """Performs various graphical filters on the image.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR
    """

    args = image.veritas.args

    # Nothing to do
    if not args.auto and not args.imageTransform:
        logger.debug('Nothing to do.')
        return

    # Set alpha if it exists.. f_a == the same image but with the alpha layer full
    bands = image.file.getbands()
    if "A" in bands:
            index = bands.index("A")
            f_a = numpy.array(image.file)
            f_a[:,:,index] = 0xff
            f_a = Image.fromarray(f_a)
    # Using an else here to keep things simple. If you use f_a you will always have a working non-alpha image
    else:
            f_a = image.file

    # Check what type of image is it
    # ColorMap
    if "P" in bands:
        logger.debug('Image is colormap, skipping.')
        return

    try:
        colorPlanes = image.file.split()
    except OSError as e:
        logger.error("%s", e)
        ImageFile.LOAD_TRUNCATED_IMAGES = True
        colorPlanes = image.file.split()

    ###########
    # Filters #
    ###########
    
    # Built-in filters to apply
    filters = [ImageFilter.EDGE_ENHANCE, ImageFilter.EDGE_ENHANCE_MORE, ImageFilter.FIND_EDGES, ImageFilter.MaxFilter, ImageFilter.MedianFilter, ImageFilter.MinFilter, ImageFilter.ModeFilter, ImageFilter.SHARPEN, ImageFilter.SMOOTH, ImageFilter.GaussianBlur]

    # {'enhancers': [(ImageEnhance.Brightness, 0.5), (ImageEnhance.Contrast, 50)], 'outname': 'enhance_contrast_50_brightness_-50.png'},
    # Each item in enhancer will be applied linearly
    enhancers = [
            {'enhancers': [(ImageEnhance.Sharpness, 25)], 'outname': 'enhance_sharpness_25.png'},
            {'enhancers': [(ImageEnhance.Sharpness, 50)], 'outname': 'enhance_sharpness_50.png'},
            {'enhancers': [(ImageEnhance.Sharpness, 75)], 'outname': 'enhance_sharpness_75.png'},
            {'enhancers': [(ImageEnhance.Sharpness, 100)], 'outname': 'enhance_sharpness_100.png'},
            {'enhancers': [(ImageEnhance.Sharpness, -25)], 'outname': 'enhance_sharpness_-25.png'},
            {'enhancers': [(ImageEnhance.Sharpness, -50)], 'outname': 'enhance_sharpness_-50.png'},
            {'enhancers': [(ImageEnhance.Sharpness, -75)], 'outname': 'enhance_sharpness_-75.png'},
            {'enhancers': [(ImageEnhance.Sharpness, -100)], 'outname': 'enhance_sharpness_-100.png'},
    ]

    image_ops = [
            {'op': ImageOps.autocontrast, 'outname': 'autocontrast.png'},
            {'op': ImageOps.grayscale, 'outname': 'grayscale.png'},
            {'op': ImageOps.equalize, 'outname': 'equalize.png'},
            {'op': ImageOps.invert, 'outname': 'inverted.png'},
            {'op': ImageOps.solarize, 'outname': 'solarized.png'},
    ]

    wait_for = []

    with multiprocessing.Pool() as pool:

        # Generic filters
        for filt in filters:
            wait_for.append(pool.apply_async(run_filter, args=(image, f_a, filt), error_callback=print_error))

        # Specific color planes
        for plane in ["red", "green", "blue", "alpha"]:
            wait_for.append(pool.apply_async(run_color_planes, args=(image, f_a, plane), error_callback=print_error))

        # Enhancers
        for run in enhancers:
            wait_for.append(pool.apply_async(run_enhancer, args=(image,), kwds=run, error_callback=print_error))
    
        # Image Ops
        for run in image_ops:
            wait_for.append(pool.apply_async(run_image_op, args=(image,), kwds=run, error_callback=print_error))

        ##################
        # Bit Map Planes #
        ##################
        
        # TODO: This is an assumption on ordering... Maybe find a better way.
        color = "Red"
        for plane in colorPlanes:
                for i in range(8):
                    # run_bit_plane(image, f_a, color, index
                    wait_for.append(pool.apply_async(run_bit_plane, args=(image, f_a, plane, color, i), error_callback=print_error))
                # Move to next color
                if color == "Red":
                        color = "Green"
                elif color == "Green":
                        color = "Blue"
                elif color == "Blue":
                        color = "Alpha"
                elif color == "Alpha":
                        color = "Error"
                else:
                        raise Exception("We got more planes than we were expecting...\nPlanes: {0}".format(colorPlanes))

        # Inverting
        #wait_for.append(pool.apply_async(run_invert, args=(image, f_a)))

        # Wait for everything to finish
        for w in wait_for: w.wait()




import logging
logger = logging.getLogger('StegoVeritas:Modules:Image:Analysis:ExtractLSB')

import os


def run(image):
    """Extracts byte stream of Least Significant Bits.

    Args:
        image: SVImage class instance

    Returns:
        None

    Saves the result to RESULTSDIR/LSBExtracted.bin
    """

    args = image.veritas.args

    # Nothing to do
    if not args.extractLSB:
        logger.debug('Nothing to do.')
        return

    if args.red:
            r = args.red
    else:
            r = []

    if args.green:
            g = args.green
    else:
            g = []

    if args.blue:
            b = args.blue
    else:
            b = []

    if args.alpha:
            a = args.alpha
    else:
            a = []

    print("Extracting ({0},{1},{2},{3})".format(r,g,b,a))

    extract_location = os.path.join(image.veritas.results_directory,"LSBExtracted.bin")
    o = image.dumpLSBRGBA(red_index=r,green_index=g,blue_index=b,alpha_index=a)
    with open(extract_location,"wb") as f:
        f.write(o)

    print("Extracted to {0}".format(extract_location))




import logging
logger = logging.getLogger('StegoVeritas:Modules:Multi:Analysis:Carve')

import os
import random

def run(multi):
    """Carve/extract anything buried in this file.

    Args:
        image: MultiHandler class instance

    Returns:
        None

    Saves the result to RESULTSDIR/keepers/
    """

    args = multi.veritas.args

    # Nothing to do
    if not args.auto and not args.carve:
        logger.debug('Nothing to do.')
        return

    # This is just a thin wrapper to hand the base file off to test_output method
    with open(multi.veritas.file_name, 'rb') as f:
        multi.veritas.test_output(f.read())
